@@ -1113,6 +1113,66 @@ int osmo_ss7_asp_peer_snprintf(char* buf, size_t buf_len, struct osmo_ss7_asp_pe
 	return len;
 }
 
+/*! \brief Set (copy) addresses for a given ASP peer. Previous addresses are freed.
+ *  \param[in] peer Application Server Process peer whose addresses are to be set.
+ *  \param[in] talloc_ctx talloc context used to allocate new addresses.
+ *  \param[in] hosts Array of strings containing IP addresses.
+ *  \param[in] host_cnt Number of strings in hosts
+ *  \returns 0 on success; negtive otherwise */
+int osmo_ss7_asp_peer_set_hosts(struct osmo_ss7_asp_peer *peer, void *talloc_ctx, const char* const* hosts, size_t host_cnt)
+{
+	int i = 0;
+
+	if (host_cnt > ARRAY_SIZE(peer->host))
+		return -EINVAL;
+
+	for (; i < host_cnt; i++)
+		osmo_talloc_replace_string(talloc_ctx, &peer->host[i], hosts[i]);
+	for (; i < peer->host_cnt; i++) {
+			talloc_free(peer->host[i]);
+			peer->host[i] = NULL;
+	}
+
+	peer->host_cnt = host_cnt;
+	return 0;
+}
+
+/*! \brief Append (copy) address to a given ASP peer. Previous addresses are kept.
+ *  \param[in] peer Application Server Process peer the address is appened to.
+ *  \param[in] talloc_ctx talloc context used to allocate new address.
+ *  \param[in] host string containing an IP addresses.
+ *  \returns 0 on success; negtive otherwise */
+int osmo_ss7_asp_peer_add_host(struct osmo_ss7_asp_peer *peer, void *talloc_ctx, const char *host)
+
+{
+	int i;
+	bool new_is_any = !host || !strcmp(host, "0.0.0.0");
+	bool iter_is_any;
+
+	if (peer->host_cnt >= ARRAY_SIZE(peer->host))
+		return -EINVAL;
+
+	/* Makes no sense to have INET_ANY many times, or INET_ANY together with
+	   specific addresses: */
+	if (new_is_any && peer->host_cnt != 0)
+		return -EINVAL;
+
+	/* Makes no sense to add specific address to set if INET_ANY is
+	   already set: */
+	if (!new_is_any) {
+		for (i = 0; i < peer->host_cnt; i++) {
+				iter_is_any = !peer->host[i] ||
+					      !strcmp(peer->host[i], "0.0.0.0");
+				if (iter_is_any)
+					return -EINVAL;
+		}
+	}
+
+	osmo_talloc_replace_string(talloc_ctx, &peer->host[peer->host_cnt], host);
+	peer->host_cnt++;
+	return 0;
+}
+
 struct osmo_ss7_asp *
 osmo_ss7_asp_find_by_name(struct osmo_ss7_instance *inst, const char *name)
 {
@@ -1726,12 +1786,13 @@ static int xua_srv_conn_closed_cb(struct osmo_stream_srv *srv)
 	/* send M-SCTP_RELEASE.ind to Layer Manager */
 	xua_asp_send_xlm_prim_simple(asp, OSMO_XLM_PRIM_M_SCTP_RELEASE, PRIM_OP_INDICATION);
 
+	asp->server = NULL;
+
 	/* if we were dynamically allocated at accept_cb() time, let's
 	 * self-destruct now.  A new connection will re-create the ASP. */
 	if (asp->dyn_allocated) {
 		/* avoid re-entrance via osmo_stream_srv_destroy() which
 		 * called us */
-		asp->server = NULL;
 		osmo_ss7_asp_destroy(asp);
 	}
 
@@ -1746,7 +1807,6 @@ static int xua_accept_cb(struct osmo_stream_srv_link *link, int fd)
 	struct osmo_xua_server *oxs = osmo_stream_srv_link_get_data(link);
 	struct osmo_stream_srv *srv;
 	struct osmo_ss7_asp *asp;
-	int i;
 	char *sock_name = osmo_sock_get_name(link, fd);
 	const char *proto_name = get_value_string(osmo_ss7_asp_protocol_vals, oxs->cfg.proto);
 
@@ -1773,6 +1833,17 @@ static int xua_accept_cb(struct osmo_stream_srv_link *link, int fd)
 	if (asp) {
 		LOGP(DLSS7, LOGL_INFO, "%s: matched connection to ASP %s\n",
 			sock_name, asp->cfg.name);
+		/* we need to check if we already have a socket associated, and close it.  Otherwise it might
+		 * happen that both the listen-fd for this accept() and the old socket are marked 'readable'
+		 * during the same scheduling interval, and we're processing them in the "wrong" order, i.e.
+		 * we first see the accept of the new fd before we see the close on the old fd */
+		if (asp->server) {
+			LOGPASP(asp, DLSS7, LOGL_FATAL, "accept of new connection from %s before old was closed "
+				"-> close old one\n", sock_name);
+			osmo_stream_srv_set_data(asp->server, NULL);
+			osmo_stream_srv_destroy(asp->server);
+			asp->server = NULL;
+		}
 	} else {
 		if (!oxs->cfg.accept_dyn_reg) {
 			LOGP(DLSS7, LOGL_NOTICE, "%s: %s connection without matching "
@@ -1786,6 +1857,7 @@ static int xua_accept_cb(struct osmo_stream_srv_link *link, int fd)
 							  oxs->cfg.proto);
 			if (asp) {
 				char hostbuf[INET6_ADDRSTRLEN];
+				const char *hostbuf_ptr = &hostbuf[0];
 				char portbuf[16];
 
 				osmo_sock_get_ip_and_port(fd, hostbuf, sizeof(hostbuf), portbuf, sizeof(portbuf), false);
@@ -1794,14 +1866,14 @@ static int xua_accept_cb(struct osmo_stream_srv_link *link, int fd)
 				asp->cfg.is_server = true;
 				asp->cfg.role = OSMO_SS7_ASP_ROLE_SG;
 				asp->cfg.local.port = oxs->cfg.local.port;
-				for (i = 0; i < oxs->cfg.local.host_cnt; i++)
-					asp->cfg.local.host[i] = talloc_strdup(asp, oxs->cfg.local.host[i]);
-				asp->cfg.local.host_cnt = oxs->cfg.local.host_cnt;
 				asp->cfg.remote.port = atoi(portbuf);
-				asp->cfg.remote.host[0] = talloc_strdup(asp, hostbuf);
-				asp->cfg.remote.host_cnt = 1;
 				asp->dyn_allocated = true;
 				asp->server = srv;
+				osmo_ss7_asp_peer_set_hosts(&asp->cfg.local, asp,
+							    (const char* const*)oxs->cfg.local.host,
+							    oxs->cfg.local.host_cnt);
+				osmo_ss7_asp_peer_set_hosts(&asp->cfg.remote, asp,
+							    &hostbuf_ptr, 1);
 				osmo_ss7_asp_restart(asp);
 			}
 		}
@@ -1810,13 +1882,13 @@ static int xua_accept_cb(struct osmo_stream_srv_link *link, int fd)
 			talloc_free(sock_name);
 			return -1;
 		}
+		llist_add_tail(&asp->siblings, &oxs->asp_list);
 	}
 
 	/* update the ASP reference back to the server over which the
 	 * connection came in */
 	asp->server = srv;
 	asp->xua_server = oxs;
-	llist_add_tail(&asp->siblings, &oxs->asp_list);
 	/* update the ASP socket name */
 	if (asp->sock_name)
 		talloc_free(asp->sock_name);
@@ -1979,58 +2051,30 @@ osmo_ss7_xua_server_bind(struct osmo_xua_server *xs)
 int
 osmo_ss7_xua_server_set_local_host(struct osmo_xua_server *xs, const char *local_host)
 {
-	osmo_ss7_xua_server_set_local_hosts(xs, &local_host, 1);
-	return 0;
+	return osmo_ss7_xua_server_set_local_hosts(xs, &local_host, 1);
 }
 
 int
 osmo_ss7_xua_server_set_local_hosts(struct osmo_xua_server *xs, const char **local_hosts, size_t local_host_cnt)
 {
-	int i = 0;
+	int rc;
 	OSMO_ASSERT(ss7_initialized);
 
-	if (local_host_cnt > ARRAY_SIZE(xs->cfg.local.host))
-		return -EINVAL;
-
-	for (; i < local_host_cnt; i++)
-		osmo_talloc_replace_string(xs, &xs->cfg.local.host[i], local_hosts[i]);
-	for (; i < xs->cfg.local.host_cnt; i++) {
-			talloc_free(xs->cfg.local.host[i]);
-			xs->cfg.local.host[i] = NULL;
-	}
-
-	xs->cfg.local.host_cnt = local_host_cnt;
-
-	osmo_stream_srv_link_set_addrs(xs->server, (const char **)xs->cfg.local.host, xs->cfg.local.host_cnt);
-
-	return 0;
+	rc = osmo_ss7_asp_peer_set_hosts(&xs->cfg.local, xs, local_hosts, local_host_cnt);
+	if (rc < 0)
+		return rc;
+	return osmo_stream_srv_link_set_addrs(xs->server, (const char **)xs->cfg.local.host, xs->cfg.local.host_cnt);
 }
 
 int
 osmo_ss7_xua_server_add_local_host(struct osmo_xua_server *xs, const char *local_host)
 {
-	int i;
-	bool new_is_any = !local_host || !strcmp(local_host, "0.0.0.0");
-	bool iter_is_any;
+	int rc;
 
-	/* Makes no sense to have INET_ANY and specific addresses in the set */
-	for (i = 0; i < xs->cfg.local.host_cnt; i++) {
-			iter_is_any = !xs->cfg.local.host[i] ||
-				      !strcmp(xs->cfg.local.host[i], "0.0.0.0");
-			if (new_is_any && iter_is_any)
-				return -EINVAL;
-			if (!new_is_any && iter_is_any)
-				return -EINVAL;
-	}
-	/* Makes no sense to have INET_ANY many times */
-	if (new_is_any && xs->cfg.local.host_cnt)
-		return -EINVAL;
-
-	osmo_talloc_replace_string(xs, &xs->cfg.local.host[xs->cfg.local.host_cnt], local_host);
-	xs->cfg.local.host_cnt++;
-
-	osmo_stream_srv_link_set_addrs(xs->server, (const char **)xs->cfg.local.host, xs->cfg.local.host_cnt);
-	return 0;
+	rc = osmo_ss7_asp_peer_add_host(&xs->cfg.local, xs, local_host);
+	if (rc < 0)
+		return rc;
+	return osmo_stream_srv_link_set_addrs(xs->server, (const char **)xs->cfg.local.host, xs->cfg.local.host_cnt);
 }
 
 void osmo_ss7_xua_server_destroy(struct osmo_xua_server *xs)
