@@ -3,6 +3,8 @@
 /* (C) 2015-2017 by Harald Welte <laforge@gnumonks.org>
  * All Rights Reserved
  *
+ * SPDX-License-Identifier: GPL-2.0+
+ *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
  * the Free Software Foundation; either version 2 of the License, or
@@ -37,6 +39,9 @@
 
 #include "xua_internal.h"
 #include <osmocom/sigtran/sccp_sap.h>
+#include "sccp_internal.h"
+
+#include "sccp_internal.h"
 
 #define XUA_VAR_STR	"(sua|m3ua|ipa)"
 
@@ -197,18 +202,23 @@ DEFUN(cs7_permit_dyn_rkm, cs7_permit_dyn_rkm_cmd,
 	return CMD_SUCCESS;
 }
 
-static void write_one_cs7(struct vty *vty, struct osmo_ss7_instance *inst);
+static void write_one_cs7(struct vty *vty, struct osmo_ss7_instance *inst, bool show_dyn_config);
 
-static int config_write_cs7(struct vty *vty)
+static int write_all_cs7(struct vty *vty, bool show_dyn_config)
 {
 	struct osmo_ss7_instance *inst;
 
 	llist_for_each_entry(inst, &osmo_ss7_instances, list)
-		write_one_cs7(vty, inst);
+		write_one_cs7(vty, inst, show_dyn_config);
 
 	return 0;
 }
 
+static int config_write_cs7(struct vty *vty)
+{
+	return write_all_cs7(vty, false);
+}
+
 DEFUN(show_cs7_user, show_cs7_user_cmd,
 	"show cs7 instance <0-15> users",
 	SHOW_STR CS7_STR INST_STR INST_STR "User Table\n")
@@ -357,7 +367,7 @@ static void write_one_rtable(struct vty *vty, struct osmo_ss7_route_table *rtabl
 	llist_for_each_entry(rt, &rtable->routes, list) {
 		vty_out(vty, "  update route %s %s linkset %s",
 			osmo_ss7_pointcode_print(rtable->inst, rt->cfg.pc),
-			osmo_ss7_pointcode_print(rtable->inst, rt->cfg.mask),
+			osmo_ss7_pointcode_print2(rtable->inst, rt->cfg.mask),
 			rt->cfg.linkset_name);
 		if (rt->cfg.priority)
 			vty_out(vty, " priority %u", rt->cfg.priority);
@@ -379,7 +389,7 @@ static void vty_dump_rtable(struct vty *vty, struct osmo_ss7_route_table *rtbl)
 
 	llist_for_each_entry(rt, &rtbl->routes, list) {
 		vty_out(vty, "%-22s %c %c %u %-19s %-7s %-7s %-7s%s",
-			osmo_ss7_pointcode_print(rtbl->inst, rt->cfg.mask),
+			osmo_ss7_route_print(rt),
 			' ', ' ', rt->cfg.priority, rt->cfg.linkset_name, "?", "?", "?", VTY_NEWLINE);
 	}
 }
@@ -431,6 +441,8 @@ DEFUN(cs7_xua, cs7_xua_cmd,
 		xs = osmo_ss7_xua_server_create(inst, proto, port, NULL);
 		if (!xs)
 			return CMD_SUCCESS;
+		/* Drop first dummy address created automatically by _create(): */
+		osmo_ss7_xua_server_set_local_hosts(xs, NULL, 0);
 	}
 
 	vty->node = L_CS7_XUA_NODE;
@@ -464,7 +476,8 @@ DEFUN(xua_local_ip, xua_local_ip_cmd,
 {
 	struct osmo_xua_server *xs = vty->index;
 
-	osmo_ss7_xua_server_set_local_host(xs, argv[0]);
+	osmo_ss7_xua_server_add_local_host(xs, argv[0]);
+
 	return CMD_SUCCESS;
 }
 
@@ -486,14 +499,63 @@ DEFUN(xua_accept_dyn_asp, xua_accept_dyn_asp_cmd,
 
 static void write_one_xua(struct vty *vty, struct osmo_xua_server *xs)
 {
+	int i;
 	vty_out(vty, " listen %s %u%s",
 		get_value_string(osmo_ss7_asp_protocol_vals, xs->cfg.proto),
 		xs->cfg.local.port, VTY_NEWLINE);
-	vty_out(vty, "  local-ip %s%s", xs->cfg.local.host, VTY_NEWLINE);
+
+	for (i = 0; i < xs->cfg.local.host_cnt; i++) {
+		if (xs->cfg.local.host[i])
+			vty_out(vty, "  local-ip %s%s", xs->cfg.local.host[i], VTY_NEWLINE);
+	}
 	if (xs->cfg.accept_dyn_reg)
 		vty_out(vty, "  accept-asp-connections dynamic-permitted%s", VTY_NEWLINE);
 }
 
+static void vty_dump_xua_server(struct vty *vty, struct osmo_xua_server *xs)
+{
+	char buf[512];
+	const char *proto = get_value_string(osmo_ss7_asp_protocol_vals, xs->cfg.proto);
+	if (osmo_ss7_asp_peer_snprintf(buf, sizeof(buf), &xs->cfg.local) < 0)
+		snprintf(buf, sizeof(buf), "<error>");
+	vty_out(vty, "xUA server for %s on %s%s", proto, buf, VTY_NEWLINE);
+}
+
+DEFUN(show_cs7_xua, show_cs7_xua_cmd,
+      "show cs7 "XUA_VAR_STR" [<0-65534>]",
+      SHOW_STR CS7_STR XUA_VAR_HELP_STR "Port Number")
+{
+	struct osmo_ss7_instance *inst;
+	struct osmo_xua_server *xs;
+	enum osmo_ss7_asp_protocol proto = parse_asp_proto(argv[0]);
+
+	llist_for_each_entry(inst, &osmo_ss7_instances, list) {
+		if (argc > 1) {
+			int port = atoi(argv[1]);
+			xs = osmo_ss7_xua_server_find(inst, proto, port);
+			if (!xs) {
+				vty_out(vty, "%% No matching server found%s", VTY_NEWLINE);
+				return CMD_WARNING;
+			}
+			vty_dump_xua_server(vty, xs);
+		} else {
+			llist_for_each_entry(xs, &inst->xua_servers, list) {
+				if (xs->cfg.proto == proto)
+					vty_dump_xua_server(vty, xs);
+			}
+		}
+	}
+	return CMD_SUCCESS;
+}
+
+DEFUN(show_cs7_config, show_cs7_config_cmd,
+      "show cs7 config",
+      SHOW_STR CS7_STR "Currently running cs7 configuration")
+{
+	write_all_cs7(vty, true);
+	return CMD_SUCCESS;
+}
+
 
 /***********************************************************************
  * Application Server Process
@@ -525,12 +587,16 @@ DEFUN(cs7_asp, cs7_asp_cmd,
 		return CMD_WARNING;
 	}
 
-	asp = osmo_ss7_asp_find_or_create(inst, name, remote_port, local_port, protocol);
+	asp = osmo_ss7_asp_find(inst, name, remote_port, local_port, protocol);
 	if (!asp) {
-		vty_out(vty, "cannot create ASP '%s'%s", name, VTY_NEWLINE);
-		return CMD_WARNING;
+		asp = osmo_ss7_asp_find_or_create(inst, name, remote_port, local_port, protocol);
+		if (!asp) {
+			vty_out(vty, "cannot create ASP '%s'%s", name, VTY_NEWLINE);
+			return CMD_WARNING;
+		}
+		asp->cfg.is_server = true;
+		asp->cfg.role = OSMO_SS7_ASP_ROLE_SG;
 	}
-	asp->cfg.is_server = true;
 
 	vty->node = L_CS7_ASP_NODE;
 	vty->index = asp;
@@ -556,13 +622,23 @@ DEFUN(no_cs7_asp, no_cs7_asp_cmd,
 	return CMD_SUCCESS;
 }
 
+DEFUN(asp_local_ip, asp_local_ip_cmd,
+	"local-ip A.B.C.D",
+	"Specify Local IP Address from which to contact ASP\n"
+	"Local IP Address from which to contact of ASP\n")
+{
+	struct osmo_ss7_asp *asp = vty->index;
+	osmo_ss7_asp_peer_add_host(&asp->cfg.local, asp, argv[0]);
+	return CMD_SUCCESS;
+}
+
 DEFUN(asp_remote_ip, asp_remote_ip_cmd,
 	"remote-ip A.B.C.D",
 	"Specify Remote IP Address of ASP\n"
 	"Remote IP Address of ASP\n")
 {
 	struct osmo_ss7_asp *asp = vty->index;
-	osmo_talloc_replace_string(asp, &asp->cfg.remote.host, argv[0]);
+	osmo_ss7_asp_peer_add_host(&asp->cfg.remote, asp, argv[0]);
 	return CMD_SUCCESS;
 }
 
@@ -576,6 +652,47 @@ DEFUN(asp_qos_clas, asp_qos_class_cmd,
 	return CMD_SUCCESS;
 }
 
+DEFUN(asp_role, asp_role_cmd,
+	"role (sg|asp|ipsp)",
+	"Specify the xUA role for this ASP\n"
+	"SG (Signaling Gateway)\n"
+	"ASP (Application Server Process)\n"
+	"IPSP (IP Signalling Point)\n")
+{
+	struct osmo_ss7_asp *asp = vty->index;
+
+	if (!strcmp(argv[0], "sg")) {
+		asp->cfg.role = OSMO_SS7_ASP_ROLE_SG;
+	} else if (!strcmp(argv[0], "asp")) {
+		asp->cfg.role = OSMO_SS7_ASP_ROLE_ASP;
+	} else if (!strcmp(argv[0], "ipsp")) {
+		vty_out(vty, "IPSP role isn't supported yet%s", VTY_NEWLINE);
+		return CMD_WARNING;
+	} else
+		OSMO_ASSERT(0);
+
+	asp->cfg.role_set_by_vty = true;
+	return CMD_SUCCESS;
+}
+
+DEFUN(sctp_role, asp_sctp_role_cmd,
+	"sctp-role (client|server)",
+	"Specify the SCTP role for this ASP\n"
+	"Operate as SCTP client; connect to a server\n"
+	"Operate as SCTP server; wait for client connections\n")
+{
+	struct osmo_ss7_asp *asp = vty->index;
+
+	if (!strcmp(argv[0], "client"))
+		asp->cfg.is_server = false;
+	else if (!strcmp(argv[0], "server"))
+		asp->cfg.is_server = true;
+	else
+		OSMO_ASSERT(0);
+
+	return CMD_SUCCESS;
+}
+
 DEFUN(asp_block, asp_block_cmd,
 	"block",
 	"Allows a SCTP Association with ASP, but doesn't let it become active\n")
@@ -600,6 +717,7 @@ DEFUN(show_cs7_asp, show_cs7_asp_cmd,
 {
 	struct osmo_ss7_instance *inst;
 	struct osmo_ss7_asp *asp;
+	char buf[512];
 	int id = atoi(argv[0]);
 
 	inst = osmo_ss7_instance_find(id);
@@ -609,23 +727,26 @@ DEFUN(show_cs7_asp, show_cs7_asp_cmd,
 	}
 
 	vty_out(vty, "                                                          Effect Primary%s", VTY_NEWLINE);
-	vty_out(vty, "ASP Name      AS Name       State          Type  Rmt Port Remote IP Addr  SCTP%s", VTY_NEWLINE);
-	vty_out(vty, "------------  ------------  -------------  ----  -------- --------------- ----------%s", VTY_NEWLINE);
+	vty_out(vty, "ASP Name      AS Name       State          Type Remote IP Addr:Rmt Port SCTP%s", VTY_NEWLINE);
+	vty_out(vty, "------------  ------------  -------------  ---- ----------------------- ----------%s", VTY_NEWLINE);
 
 	llist_for_each_entry(asp, &inst->asp_list, list) {
-		vty_out(vty, "%-12s  %-12s  %-13s  %-4s  %-8u %-15s %-10s%s",
-			asp->cfg.name, "?", osmo_fsm_inst_state_name(asp->fi),
+		osmo_ss7_asp_peer_snprintf(buf, sizeof(buf), &asp->cfg.remote);
+		vty_out(vty, "%-12s  %-12s  %-13s  %-4s  %-14s  %-10s%s",
+			asp->cfg.name, "?",
+			asp->fi? osmo_fsm_inst_state_name(asp->fi) : "uninitialized",
 			get_value_string(osmo_ss7_asp_protocol_vals, asp->cfg.proto),
-			asp->cfg.remote.port, asp->cfg.remote.host, "", VTY_NEWLINE);
+			buf, "", VTY_NEWLINE);
 	}
 	return CMD_SUCCESS;
 }
 
-static void write_one_asp(struct vty *vty, struct osmo_ss7_asp *asp)
+static void write_one_asp(struct vty *vty, struct osmo_ss7_asp *asp, bool show_dyn_config)
 {
-	/* skip any dynamically created ASPs (auto-created at connect
-	 * time) */
-	if (asp->dyn_allocated)
+	int i;
+	/* skip any dynamically created ASPs (e.g. auto-created at connect time) */
+	if ((asp->dyn_allocated || asp->simple_client_allocated)
+	    && !show_dyn_config)
 		return;
 
 	vty_out(vty, " asp %s %u %u %s%s",
@@ -633,10 +754,22 @@ static void write_one_asp(struct vty *vty, struct osmo_ss7_asp *asp)
 		osmo_ss7_asp_protocol_name(asp->cfg.proto), VTY_NEWLINE);
 	if (asp->cfg.description)
 		vty_out(vty, "  description %s%s", asp->cfg.description, VTY_NEWLINE);
-	if (asp->cfg.remote.host)
-		vty_out(vty, "  remote-ip %s%s", asp->cfg.remote.host, VTY_NEWLINE);
+	for (i = 0; i < asp->cfg.local.host_cnt; i++) {
+		if (asp->cfg.local.host[i])
+			vty_out(vty, "  local-ip %s%s", asp->cfg.local.host[i], VTY_NEWLINE);
+	}
+	for (i = 0; i < asp->cfg.remote.host_cnt; i++) {
+		if (asp->cfg.remote.host[i])
+			vty_out(vty, "  remote-ip %s%s", asp->cfg.remote.host[i], VTY_NEWLINE);
+	}
 	if (asp->cfg.qos_class)
 		vty_out(vty, "  qos-class %u%s", asp->cfg.qos_class, VTY_NEWLINE);
+	if (asp->cfg.role_set_by_vty) {
+		vty_out(vty, "  role %s%s", osmo_str_tolower(get_value_string(osmo_ss7_asp_role_names, asp->cfg.role)),
+			VTY_NEWLINE);
+	}
+	if (!asp->cfg.is_server)
+		vty_out(vty, "  sctp-role client%s", VTY_NEWLINE);
 }
 
 
@@ -662,7 +795,7 @@ DEFUN(cs7_as, cs7_as_cmd,
 	enum osmo_ss7_asp_protocol protocol = parse_asp_proto(argv[1]);
 
 	if (protocol == OSMO_SS7_ASP_PROT_NONE) {
-		vty_out(vty, "invalid protocol '%s'%s", argv[3], VTY_NEWLINE);
+		vty_out(vty, "invalid protocol '%s'%s", argv[1], VTY_NEWLINE);
 		return CMD_WARNING;
 	}
 
@@ -741,6 +874,18 @@ DEFUN(as_traf_mode, as_traf_mode_cmd,
 	struct osmo_ss7_as *as = vty->index;
 
 	as->cfg.mode = get_string_value(osmo_ss7_as_traffic_mode_vals, argv[0]);
+	as->cfg.mode_set_by_vty = true;
+	return CMD_SUCCESS;
+}
+
+DEFUN(as_no_traf_mode, as_no_traf_mode_cmd,
+	"no traffic-mode",
+	NO_STR "Remove explicit traffic mode of operation of this AS\n")
+{
+	struct osmo_ss7_as *as = vty->index;
+
+	as->cfg.mode = 0;
+	as->cfg.mode_set_by_vty = false;
 	return CMD_SUCCESS;
 }
 
@@ -807,6 +952,12 @@ static int _rout_key(struct vty *vty,
 	struct osmo_ss7_routing_key *rkey = &as->cfg.routing_key;
 	int pc;
 
+	if (as->cfg.proto == OSMO_SS7_ASP_PROT_IPA && atoi(rcontext) != 0) {
+		vty_out(vty, "IPA doesn't support routing contexts; only permitted routing context "
+			"is 0\n");
+		return CMD_WARNING;
+	}
+
 	pc = osmo_ss7_pointcode_parse(as->inst, dpc);
 	if (pc < 0) {
 		vty_out(vty, "Invalid point code (%s)%s", dpc, VTY_NEWLINE);
@@ -875,13 +1026,38 @@ DEFUN(as_pc_override, as_pc_override_cmd,
 	return CMD_SUCCESS;
 }
 
-static void write_one_as(struct vty *vty, struct osmo_ss7_as *as)
+DEFUN(as_pc_patch_sccp, as_pc_patch_sccp_cmd,
+	"point-code override patch-sccp (disabled|both)",
+	"Point Code Specific Features\n"
+	"Override (force) a point-code to hard-coded value\n"
+	"Patch point code values into SCCP called/calling address\n"
+	"Don't patch any point codes into SCCP called/calling address\n"
+	"Patch both origin and destination point codes into SCCP called/calling address\n")
+{
+	struct osmo_ss7_as *as = vty->index;
+
+	if (as->cfg.proto != OSMO_SS7_ASP_PROT_IPA) {
+		vty_out(vty, "Only IPA type AS support point-code patch-into-sccp. "
+			"Be happy that you don't need it!%s", VTY_NEWLINE);
+		return CMD_WARNING;
+	}
+
+	if (!strcmp(argv[0], "disabled"))
+		as->cfg.pc_override.sccp_mode = OSMO_SS7_PATCH_NONE;
+	else
+		as->cfg.pc_override.sccp_mode = OSMO_SS7_PATCH_BOTH;
+
+	return CMD_SUCCESS;
+}
+
+static void write_one_as(struct vty *vty, struct osmo_ss7_as *as, bool show_dyn_config)
 {
 	struct osmo_ss7_routing_key *rkey;
 	unsigned int i;
 
 	/* skip any dynamically allocated AS definitions */
-	if (as->rkm_dyn_allocated)
+	if ((as->rkm_dyn_allocated || as->simple_client_allocated)
+	    && !show_dyn_config)
 		return;
 
 	vty_out(vty, " as %s %s%s", as->cfg.name,
@@ -892,9 +1068,13 @@ static void write_one_as(struct vty *vty, struct osmo_ss7_as *as)
 		struct osmo_ss7_asp *asp = as->cfg.asps[i];
 		if (!asp)
 			continue;
+		/* skip any dynamically created ASPs (e.g. auto-created at connect time) */
+		if ((asp->dyn_allocated || asp->simple_client_allocated)
+		    && !show_dyn_config)
+			continue;
 		vty_out(vty, "  asp %s%s", asp->cfg.name, VTY_NEWLINE);
 	}
-	if (as->cfg.mode != OSMO_SS7_AS_TMOD_LOADSHARE)
+	if (as->cfg.mode_set_by_vty)
 		vty_out(vty, "  traffic-mode %s%s",
 			osmo_ss7_as_traffic_mode_name(as->cfg.mode), VTY_NEWLINE);
 	if (as->cfg.recovery_timeout_msec != 2000) {
@@ -914,8 +1094,11 @@ static void write_one_as(struct vty *vty, struct osmo_ss7_as *as)
 	vty_out(vty, "%s", VTY_NEWLINE);
 
 	if (as->cfg.pc_override.dpc)
-		vty_out(vty, " point-code override dpc %s%s",
+		vty_out(vty, "  point-code override dpc %s%s",
 			osmo_ss7_pointcode_print(as->inst, as->cfg.pc_override.dpc), VTY_NEWLINE);
+
+	if (as->cfg.pc_override.sccp_mode)
+		vty_out(vty, "  point-code override patch-sccp both%s", VTY_NEWLINE);
 }
 
 DEFUN(show_cs7_as, show_cs7_as_cmd,
@@ -937,20 +1120,22 @@ DEFUN(show_cs7_as, show_cs7_as_cmd,
 		return CMD_WARNING;
 	}
 
-	vty_out(vty, "                          Routing    Routing Key                          Cic   Cic%s", VTY_NEWLINE);
-	vty_out(vty, "AS Name      State        Context    Dpc           Si   Opc           Ssn Min   Max%s", VTY_NEWLINE);
-	vty_out(vty, "------------ ------------ ---------- ------------- ---- ------------- --- ----- -----%s", VTY_NEWLINE);
+	vty_out(vty, "                          Routing    Routing Key                          Cic   Cic   Traffic%s", VTY_NEWLINE);
+	vty_out(vty, "AS Name      State        Context    Dpc           Si   Opc           Ssn Min   Max   Mode%s", VTY_NEWLINE);
+	vty_out(vty, "------------ ------------ ---------- ------------- ---- ------------- --- ----- ----- -------%s", VTY_NEWLINE);
 
 	llist_for_each_entry(as, &inst->as_list, list) {
 		if (filter && !strcmp(filter, "m3ua") && as->cfg.proto != OSMO_SS7_ASP_PROT_M3UA)
 			continue;
 		if (filter && !strcmp(filter, "sua") && as->cfg.proto != OSMO_SS7_ASP_PROT_SUA)
 			continue;
-		/* FIXME: active filter */
-		vty_out(vty, "%-12s %-12s %-10u %-13s %4s %13s %3s %5s %4s%s",
+		if (filter && !strcmp(filter, "active") && !osmo_ss7_as_active(as))
+			continue;
+		vty_out(vty, "%-12s %-12s %-10u %-13s %4s %13s %3s %5s %4s %10s%s",
 			as->cfg.name, osmo_fsm_inst_state_name(as->fi), as->cfg.routing_key.context,
 			osmo_ss7_pointcode_print(as->inst, as->cfg.routing_key.pc),
-			"", "", "", "", "", VTY_NEWLINE);
+			"", "", "", "", "", osmo_ss7_as_traffic_mode_name(as->cfg.mode),
+			VTY_NEWLINE);
 	}
 	return CMD_SUCCESS;
 }
@@ -964,7 +1149,7 @@ struct osmo_sccp_addr_entry {
 	struct llist_head list;
 	struct llist_head list_global;
 	struct osmo_ss7_instance *inst;
-	char name[512];
+	char name[32];
 	struct osmo_sccp_addr addr;
 };
 
@@ -1109,8 +1294,9 @@ static void write_sccp_addressbook(struct vty *vty,
 
 /* List all addressbook entries */
 DEFUN(cs7_show_sccpaddr, cs7_show_sccpaddr_cmd,
-      "show cs7 instance <0-15> sccp-addressbook",
-      SHOW_STR CS7_STR INST_STR INST_STR "List all SCCP addressbook entries\n")
+      "show cs7 instance <0-15> sccp addressbook",
+      SHOW_STR CS7_STR INST_STR INST_STR SCCP_STR
+      "List all SCCP addressbook entries\n")
 {
 	struct osmo_ss7_instance *inst;
 	struct osmo_sccp_addr_entry *entry;
@@ -1234,8 +1420,9 @@ DEFUN(cs7_sccpaddr, cs7_sccpaddr_cmd,
 	const char *name = argv[0];
 
 	if (strlen(name) >= sizeof(entry->name)) {
-		vty_out(vty, "sccp address name to long!%s", VTY_NEWLINE);
-		return CMD_WARNING;
+		vty_out(vty, "Error: SCCP address name too long: '%s'%s",
+			name, VTY_NEWLINE);
+		return CMD_ERR_INCOMPLETE;
 	}
 
 	/* Ensure that we do not use address names that
@@ -1243,9 +1430,9 @@ DEFUN(cs7_sccpaddr, cs7_sccpaddr_cmd,
 	entry = addr_entry_by_name_global(name);
 	if (entry != NULL) {
 		vty_out(vty,
-			"address name (%s) already used in ss7 instance %u%s",
-			entry->name, entry->inst->cfg.id, VTY_NEWLINE);
-		return CMD_WARNING;
+			"Error: SCCP address name already used in cs7 instance %u: '%s'%s",
+			entry->inst->cfg.id, entry->name, VTY_NEWLINE);
+		return CMD_ERR_INCOMPLETE;
 	}
 
 	entry = addr_entry_by_name_local(name, inst);
@@ -1540,7 +1727,7 @@ DEFUN(cs7_sccpaddr_gt_digits, cs7_sccpaddr_gt_digits_cmd,
  * Common
  ***********************************************************************/
 
-static void write_one_cs7(struct vty *vty, struct osmo_ss7_instance *inst)
+static void write_one_cs7(struct vty *vty, struct osmo_ss7_instance *inst, bool show_dyn_config)
 {
 	struct osmo_ss7_asp *asp;
 	struct osmo_ss7_as *as;
@@ -1581,11 +1768,11 @@ static void write_one_cs7(struct vty *vty, struct osmo_ss7_instance *inst)
 
 	/* first dump ASPs, as ASs reference them */
 	llist_for_each_entry(asp, &inst->asp_list, list)
-		write_one_asp(vty, asp);
+		write_one_asp(vty, asp, show_dyn_config);
 
 	/* then dump ASPs, as routes reference them */
 	llist_for_each_entry(as, &inst->as_list, list)
-		write_one_as(vty, as);
+		write_one_as(vty, as, show_dyn_config);
 
 	/* now dump everything that is relevent for the SG role */
 	if (cs7_role == CS7_ROLE_SG) {
@@ -1600,6 +1787,9 @@ static void write_one_cs7(struct vty *vty, struct osmo_ss7_instance *inst)
 
 	/* Append SCCP Addressbook */
 	write_sccp_addressbook(vty, inst);
+
+	if (inst->sccp)
+		osmo_sccp_vty_write_cs7_node(vty, " ", inst->sccp);
 }
 
 
@@ -1614,6 +1804,12 @@ int osmo_ss7_vty_go_parent(struct vty *vty)
 	switch (vty->node) {
 	case L_CS7_ASP_NODE:
 		asp = vty->index;
+		/* If no local addr was set */
+		if (!asp->cfg.local.host_cnt)
+			osmo_ss7_asp_peer_add_host(&asp->cfg.local, asp, NULL);
+		/* If no remote addr was set */
+		if (!asp->cfg.remote.host_cnt)
+			osmo_ss7_asp_peer_add_host(&asp->cfg.remote, asp, "127.0.0.1");
 		osmo_ss7_asp_restart(asp);
 		vty->node = L_CS7_NODE;
 		vty->index = asp->inst;
@@ -1630,6 +1826,11 @@ int osmo_ss7_vty_go_parent(struct vty *vty)
 		break;
 	case L_CS7_XUA_NODE:
 		oxs = vty->index;
+		/* If no local addr was set, or erased after _create(): */
+		if (!oxs->cfg.local.host_cnt)
+			osmo_ss7_xua_server_set_local_host(oxs, NULL);
+		if (osmo_ss7_xua_server_bind(oxs) < 0)
+			vty_out(vty, "%% Unable to bind xUA server to IP(s)%s", VTY_NEWLINE);
 		vty->node = L_CS7_NODE;
 		vty->index = oxs->inst;
 		break;
@@ -1640,10 +1841,12 @@ int osmo_ss7_vty_go_parent(struct vty *vty)
 		break;
 	case L_CS7_SCCPADDR_GT_NODE:
 		vty->node = L_CS7_SCCPADDR_NODE;
+		vty->index = NULL;
 		break;
 	case L_CS7_NODE:
 	default:
 		vty->node = CONFIG_NODE;
+		vty->index = NULL;
 		break;
 	}
 	return 0;
@@ -1669,8 +1872,7 @@ int osmo_ss7_is_config_node(struct vty *vty, int node)
 static void vty_init_addr(void)
 {
 	install_node(&sccpaddr_node, NULL);
-	vty_install_default(L_CS7_SCCPADDR_NODE);
-	install_element(L_CS7_NODE, &cs7_show_sccpaddr_cmd);
+	install_element_ve(&cs7_show_sccpaddr_cmd);
 	install_element(L_CS7_NODE, &cs7_sccpaddr_cmd);
 	install_element(L_CS7_NODE, &cs7_sccpaddr_del_cmd);
 	install_element(L_CS7_SCCPADDR_NODE, &cs7_sccpaddr_pc_del_cmd);
@@ -1692,7 +1894,6 @@ static void vty_init_addr(void)
 #endif
 	install_element(L_CS7_SCCPADDR_NODE, &cs7_sccpaddr_gt_cmd);
 	install_node(&sccpaddr_gt_node, NULL);
-	vty_install_default(L_CS7_SCCPADDR_GT_NODE);
 	install_element(L_CS7_SCCPADDR_GT_NODE, &cs7_sccpaddr_gt_gti_cmd);
 	install_element(L_CS7_SCCPADDR_GT_NODE, &cs7_sccpaddr_gt_tt_cmd);
 	install_element(L_CS7_SCCPADDR_GT_NODE, &cs7_sccpaddr_gt_npi_cmd);
@@ -1705,12 +1906,13 @@ static void vty_init_shared(void *ctx)
 	g_ctx = ctx;
 
 	install_element_ve(&show_cs7_user_cmd);
+	install_element_ve(&show_cs7_xua_cmd);
+	install_element_ve(&show_cs7_config_cmd);
 
 	/* the mother of all VTY config nodes */
 	install_element(CONFIG_NODE, &cs7_instance_cmd);
 
 	install_node(&cs7_node, config_write_cs7);
-	vty_install_default(L_CS7_NODE);
 	install_element(L_CS7_NODE, &cfg_description_cmd);
 	install_element(L_CS7_NODE, &cs7_net_ind_cmd);
 	install_element(L_CS7_NODE, &cs7_point_code_cmd);
@@ -1720,18 +1922,19 @@ static void vty_init_shared(void *ctx)
 	install_element(L_CS7_NODE, &cs7_permit_dyn_rkm_cmd);
 
 	install_node(&asp_node, NULL);
-	vty_install_default(L_CS7_ASP_NODE);
 	install_element_ve(&show_cs7_asp_cmd);
 	install_element(L_CS7_NODE, &cs7_asp_cmd);
 	install_element(L_CS7_NODE, &no_cs7_asp_cmd);
 	install_element(L_CS7_ASP_NODE, &cfg_description_cmd);
 	install_element(L_CS7_ASP_NODE, &asp_remote_ip_cmd);
+	install_element(L_CS7_ASP_NODE, &asp_local_ip_cmd);
 	install_element(L_CS7_ASP_NODE, &asp_qos_class_cmd);
+	install_element(L_CS7_ASP_NODE, &asp_role_cmd);
+	install_element(L_CS7_ASP_NODE, &asp_sctp_role_cmd);
 	install_element(L_CS7_ASP_NODE, &asp_block_cmd);
 	install_element(L_CS7_ASP_NODE, &asp_shutdown_cmd);
 
 	install_node(&as_node, NULL);
-	vty_install_default(L_CS7_AS_NODE);
 	install_element_ve(&show_cs7_as_cmd);
 	install_element(L_CS7_NODE, &cs7_as_cmd);
 	install_element(L_CS7_NODE, &no_cs7_as_cmd);
@@ -1739,6 +1942,7 @@ static void vty_init_shared(void *ctx)
 	install_element(L_CS7_AS_NODE, &as_asp_cmd);
 	install_element(L_CS7_AS_NODE, &as_no_asp_cmd);
 	install_element(L_CS7_AS_NODE, &as_traf_mode_cmd);
+	install_element(L_CS7_AS_NODE, &as_no_traf_mode_cmd);
 	install_element(L_CS7_AS_NODE, &as_recov_tout_cmd);
 	install_element(L_CS7_AS_NODE, &as_qos_class_cmd);
 	install_element(L_CS7_AS_NODE, &as_rout_key_cmd);
@@ -1746,6 +1950,7 @@ static void vty_init_shared(void *ctx)
 	install_element(L_CS7_AS_NODE, &as_rout_key_ssn_cmd);
 	install_element(L_CS7_AS_NODE, &as_rout_key_si_ssn_cmd);
 	install_element(L_CS7_AS_NODE, &as_pc_override_cmd);
+	install_element(L_CS7_AS_NODE, &as_pc_patch_sccp_cmd);
 
 	vty_init_addr();
 }
@@ -1762,7 +1967,6 @@ void osmo_ss7_vty_init_sg(void *ctx)
 	vty_init_shared(ctx);
 
 	install_node(&rtable_node, NULL);
-	vty_install_default(L_CS7_RTABLE_NODE);
 	install_element_ve(&show_cs7_route_cmd);
 	install_element(L_CS7_NODE, &cs7_route_table_cmd);
 	install_element(L_CS7_RTABLE_NODE, &cfg_description_cmd);
@@ -1770,7 +1974,6 @@ void osmo_ss7_vty_init_sg(void *ctx)
 	install_element(L_CS7_RTABLE_NODE, &cs7_rt_rem_cmd);
 
 	install_node(&xua_node, NULL);
-	vty_install_default(L_CS7_XUA_NODE);
 	install_element(L_CS7_NODE, &cs7_xua_cmd);
 	install_element(L_CS7_NODE, &no_cs7_xua_cmd);
 	install_element(L_CS7_XUA_NODE, &xua_local_ip_cmd);
