@@ -3,6 +3,8 @@
 /* (C) 2017 by Harald Welte <laforge@gnumonks.org>
  * All Rights Reserved
  *
+ * SPDX-License-Identifier: GPL-2.0+
+ *
  * based on my 2011 Erlang implementation osmo_ss7/src/sua_sccp_conv.erl
  *
  * References: ITU-T Q.713 and IETF RFC 3868
@@ -23,6 +25,7 @@
 
 #include <stdbool.h>
 #include <string.h>
+#include <limits.h>
 
 #include <osmocom/core/linuxlist.h>
 #include <osmocom/core/logging.h>
@@ -37,6 +40,8 @@
 #include "xua_internal.h"
 
 /*! \brief Find a SCCP User registered for given PC+SSN or SSN only
+ * First search all users with a valid PC for a full PC+SSN match.
+ * If no such match was found, search all users with an invalid PC for an SSN-only match.
  *  \param[in] inst SCCP Instance in which to search
  *  \param[in] ssn Sub-System Number to search for
  *  \param[in] pc Point Code to search for
@@ -46,10 +51,12 @@ sccp_user_find(struct osmo_sccp_instance *inst, uint16_t ssn, uint32_t pc)
 {
 	struct osmo_sccp_user *scu;
 
-	/* First try to find match for PC + SSN */
-	llist_for_each_entry(scu, &inst->users, list) {
-		if (osmo_ss7_pc_is_valid(scu->pc) && scu->pc == pc && scu->ssn == ssn)
-			return scu;
+	if (osmo_ss7_pc_is_valid(pc)) {
+		/* First try to find match for PC + SSN */
+		llist_for_each_entry(scu, &inst->users, list) {
+			if (osmo_ss7_pc_is_valid(scu->pc) && scu->pc == pc && scu->ssn == ssn)
+				return scu;
+		}
 	}
 
 	/* Then try to match on SSN only */
@@ -61,6 +68,20 @@ sccp_user_find(struct osmo_sccp_instance *inst, uint16_t ssn, uint32_t pc)
 	return NULL;
 }
 
+/*! Find a SCCP User registered for given PC+SSN or SSN only.
+ * First search all users with a valid PC for a full PC+SSN match.
+ * If no match was found, search all users with an invalid PC for an SSN-only match.
+ *  \param[in] inst SCCP Instance in which to search.
+ *  \param[in] ssn Sub-System Number to search for.
+ *  \param[in] pc Point Code to search for.
+ *  \returns Matching SCCP User; NULL if none found.
+ */
+struct osmo_sccp_user *
+osmo_sccp_user_find(struct osmo_sccp_instance *inst, uint16_t ssn, uint32_t pc)
+{
+	return sccp_user_find(inst, ssn, pc);
+}
+
 /*! \brief Bind a SCCP User to a given Point Code
  *  \param[in] inst SCCP Instance
  *  \param[in] name human-readable name
@@ -169,9 +190,16 @@ static int mtp_user_prim_cb(struct osmo_prim_hdr *oph, void *ctx)
 	case OSMO_PRIM(OSMO_MTP_PRIM_TRANSFER, PRIM_OP_INDICATION):
 		/* Convert from SCCP to SUA in xua_msg format */
 		xua = osmo_sccp_to_xua(oph->msg);
+		if (!xua) {
+			LOGP(DLSCCP, LOGL_ERROR, "Couldn't convert SCCP to SUA: %s\n",
+				msgb_hexdump(oph->msg));
+			rc = -1;
+			break;
+		}
 		xua->mtp = omp->u.transfer;
 		/* hand this primitive into SCCP via the SCRC code */
 		rc = scrc_rx_mtp_xfer_ind_xua(inst, xua);
+		xua_msg_free(xua);
 		break;
 	default:
 		LOGP(DLSCCP, LOGL_ERROR, "Unknown primitive %u:%u receivd\n",
@@ -291,6 +319,145 @@ bool osmo_sccp_check_addr(struct osmo_sccp_addr *addr, uint32_t presence)
 	return true;
 }
 
+/*! Compare two SCCP Global Titles.
+ * \param[in] a  left side.
+ * \param[in] b  right side.
+ * \return -1 if a < b, 1 if a > b, and 0 if a == b.
+ */
+int osmo_sccp_gt_cmp(const struct osmo_sccp_gt *a, const struct osmo_sccp_gt *b)
+{
+	if (a == b)
+		return 0;
+	if (!a)
+		return -1;
+	if (!b)
+		return 1;
+	return memcmp(a, b, sizeof(*a));
+}
+
+/*! Compare two SCCP addresses by given presence criteria.
+ * Any OSMO_SCCP_ADDR_T_* type not set in presence_criteria is ignored.
+ * In case all bits are set in presence_criteria, the comparison is in the order of:
+ * OSMO_SCCP_ADDR_T_GT, OSMO_SCCP_ADDR_T_PC, OSMO_SCCP_ADDR_T_IPv4, OSMO_SCCP_ADDR_T_IPv6, OSMO_SCCP_ADDR_T_SSN.
+ * The SCCP addresses' Routing Indicator is not compared, see osmo_sccp_addr_ri_cmp().
+ * \param[in] a  left side.
+ * \param[in] b  right side.
+ * \param[in] presence_criteria  A bitmask of OSMO_SCCP_ADDR_T_* values, or 0xffffffff to compare all parts, except the
+ *                               routing indicator.
+ * \return -1 if a < b, 1 if a > b, and 0 if all checked values match.
+ */
+int osmo_sccp_addr_cmp(const struct osmo_sccp_addr *a, const struct osmo_sccp_addr *b, uint32_t presence_criteria)
+{
+	int rc;
+	if (a == b)
+		return 0;
+	if (!a)
+		return -1;
+	if (!b)
+		return 1;
+
+	if (presence_criteria & OSMO_SCCP_ADDR_T_GT) {
+		if ((a->presence & OSMO_SCCP_ADDR_T_GT) != (b->presence & OSMO_SCCP_ADDR_T_GT))
+			return (b->presence & OSMO_SCCP_ADDR_T_GT) ? -1 : 1;
+		rc = osmo_sccp_gt_cmp(&a->gt, &b->gt);
+		if (rc)
+			return rc;
+	}
+
+	if (presence_criteria & OSMO_SCCP_ADDR_T_PC) {
+		if ((a->presence & OSMO_SCCP_ADDR_T_PC) != (b->presence & OSMO_SCCP_ADDR_T_PC))
+			return (b->presence & OSMO_SCCP_ADDR_T_PC) ? -1 : 1;
+
+		if ((a->presence & OSMO_SCCP_ADDR_T_PC)
+		    && a->pc != b->pc)
+			return (a->pc < b->pc)? -1 : 1;
+	}
+
+	if (presence_criteria & OSMO_SCCP_ADDR_T_IPv4) {
+		if ((a->presence & OSMO_SCCP_ADDR_T_IPv4) != (b->presence & OSMO_SCCP_ADDR_T_IPv4))
+			return (b->presence & OSMO_SCCP_ADDR_T_IPv4) ? -1 : 1;
+		rc = memcmp(&a->ip.v4, &b->ip.v4, sizeof(a->ip.v4));
+		if (rc)
+			return rc;
+	}
+
+	if (presence_criteria & OSMO_SCCP_ADDR_T_IPv6) {
+		if ((a->presence & OSMO_SCCP_ADDR_T_IPv6) != (b->presence & OSMO_SCCP_ADDR_T_IPv6))
+			return (b->presence & OSMO_SCCP_ADDR_T_IPv6) ? -1 : 1;
+		rc = memcmp(&a->ip.v6, &b->ip.v6, sizeof(a->ip.v6));
+		if (rc)
+			return rc;
+	}
+
+	if (presence_criteria & OSMO_SCCP_ADDR_T_SSN) {
+		if ((a->presence & OSMO_SCCP_ADDR_T_SSN) != (b->presence & OSMO_SCCP_ADDR_T_SSN))
+			return (b->presence & OSMO_SCCP_ADDR_T_SSN) ? -1 : 1;
+		if (a->ssn != b->ssn)
+			return (a->ssn < b->ssn) ? -1 : 1;
+	}
+
+	return 0;
+}
+
+/*! Compare the routing information of two SCCP addresses.
+ * Compare the ri of a and b, and, if equal, return osmo_sccp_addr_cmp() with presence criteria selected according to
+ * ri.
+ * \param[in] a  left side.
+ * \param[in] b  right side.
+ * \return -1 if a < b, 1 if a > b, and 0 if a == b.
+ */
+int osmo_sccp_addr_ri_cmp(const struct osmo_sccp_addr *a, const struct osmo_sccp_addr *b)
+{
+	uint32_t presence_criteria;
+	if (a == b)
+		return 0;
+	if (!a)
+		return -1;
+	if (!b)
+		return 1;
+	if (a->ri != b->ri)
+		return (a->ri < b->ri) ? -1 : 1;
+	switch (a->ri) {
+	case OSMO_SCCP_RI_NONE:
+		return 0;
+	case OSMO_SCCP_RI_GT:
+		presence_criteria = OSMO_SCCP_ADDR_T_GT;
+		break;
+	case OSMO_SCCP_RI_SSN_PC:
+		presence_criteria = OSMO_SCCP_ADDR_T_SSN | OSMO_SCCP_ADDR_T_PC;
+		break;
+	case OSMO_SCCP_RI_SSN_IP:
+		/* Pick IPv4 or v6 depending on what a->presence indicates. */
+		presence_criteria = OSMO_SCCP_ADDR_T_SSN | (a->presence & (OSMO_SCCP_ADDR_T_IPv4 | OSMO_SCCP_ADDR_T_IPv6));
+		break;
+	default:
+		return 0;
+	}
+
+	return osmo_sccp_addr_cmp(a, b, presence_criteria);
+}
+
+/*! Compose a human readable string to describe the SCCP user's connection.
+ * The output follows ['<scu.name>':]<local-sccp-addr>, e.g.  "'OsmoHNBW':RI=SSN_PC,PC=0.23.5,SSN=RANAP",
+ * or just "RI=SSN_PC,PC=0.23.5,SSN=RANAP" if no scu->name is set.
+ * This calls osmo_sccp_addr_name(), which returns a static buffer; hence calling this function and
+ * osmo_sccp_addr_name() in the same printf statement is likely to conflict. */
+const char *osmo_sccp_user_name(struct osmo_sccp_user *scu)
+{
+	static char buf[128];
+	struct osmo_sccp_addr sca;
+	/* Interestingly enough, the osmo_sccp_user stores an SSN and PC, but not in an osmo_sccp_addr
+	 * struct. To be able to use osmo_sccp_addr_name(), we need to first create an osmo_sccp_addr. */
+	osmo_sccp_make_addr_pc_ssn(&sca, scu->pc, scu->ssn);
+	snprintf(buf, sizeof(buf),
+		 "%s%s%s",
+		 scu->name && *scu->name ? scu->name : "",
+		 scu->name && *scu->name ? ":" : "",
+		 osmo_sccp_addr_name(scu->inst->ss7, &sca));
+	buf[sizeof(buf)-1] = '\0';
+	return buf;
+}
+
 /***********************************************************************
  * Convenience function for CLIENT
  ***********************************************************************/
@@ -373,13 +540,6 @@ osmo_sccp_simple_client_on_ss7_id(void *ctx, uint32_t ss7_id, const char *name,
 	LOGP(DLSCCP, LOGL_NOTICE, "%s: Using SS7 instance %u, pc:%s\n", name,
 	     ss7->cfg.id, osmo_ss7_pointcode_print(ss7, ss7->cfg.primary_pc));
 
-	/* There must not be an existing SCCP istance, regarless if the simple
-	 * client has created the SS7 instance or if it was already present.
-	 * An already existing SCCP instance would be an indication that this
-	 * function has been called twice with the same SS7 instance, which
-	 * must not be the case! */
-	OSMO_ASSERT(ss7->sccp == NULL);
-
 	/* Check if there is already an application server that matches
 	 * the protocol we intend to use. If not, we will create one. */
 	as = osmo_ss7_as_find_by_proto(ss7, prot);
@@ -393,6 +553,7 @@ osmo_sccp_simple_client_on_ss7_id(void *ctx, uint32_t ss7_id, const char *name,
 			goto out_ss7;
 		as_created = true;
 		as->cfg.routing_key.pc = ss7->cfg.primary_pc;
+		as->simple_client_allocated = true;
 	}
 	LOGP(DLSCCP, LOGL_NOTICE, "%s: Using AS instance %s\n", name,
 	     as->cfg.name);
@@ -414,26 +575,53 @@ osmo_sccp_simple_client_on_ss7_id(void *ctx, uint32_t ss7_id, const char *name,
 	 * we intend to use. */
 	asp = osmo_ss7_asp_find_by_proto(as, prot);
 	if (!asp) {
-		LOGP(DLSCCP, LOGL_NOTICE, "%s: Creating ASP instance\n",
-		     name);
-		asp_name = talloc_asprintf(ctx, "asp-clnt-%s", name);
-		asp =
-		    osmo_ss7_asp_find_or_create(ss7, asp_name,
-						default_remote_port,
-						default_local_port, prot);
-		talloc_free(asp_name);
-		if (!asp)
-			goto out_rt;
-		asp_created = true;
-
-		asp->cfg.local.host = default_local_ip ? talloc_strdup(asp, default_local_ip) : NULL;
-		asp->cfg.remote.host = default_remote_ip ? talloc_strdup(asp, default_remote_ip) : NULL;
+		/* Check if the user has created an ASP for this proto that is not added on any AS yet. */
+		struct osmo_ss7_asp *asp_i;
+		llist_for_each_entry(asp_i, &ss7->asp_list, list) {
+			struct osmo_ss7_as *as_i;
+			bool is_on_as = false;
+			if (asp_i->cfg.proto != prot)
+				continue;
+			llist_for_each_entry(as_i, &ss7->as_list, list) {
+				if (!osmo_ss7_as_has_asp(as_i, asp_i))
+					continue;
+				is_on_as = true;
+				break;
+			}
+			if (is_on_as) {
+				/* This ASP is already on another AS. If it was on this AS, we'd have found it above. */
+				continue;
+			}
+			/* This ASP matches the protocol and is not yet associated to any AS. Use it. */
+			asp = asp_i;
+			LOGP(DLSCCP, LOGL_NOTICE, "%s: ASP %s for %s is not associated with any AS, using it\n",
+			     name, asp->cfg.name, osmo_ss7_asp_protocol_name(prot));
+			break;
+		}
+		if (!asp) {
+			asp_name = talloc_asprintf(ctx, "asp-clnt-%s", name);
+			LOGP(DLSCCP, LOGL_NOTICE, "%s: No unassociated ASP for %s, creating new ASP %s\n",
+			     name, osmo_ss7_asp_protocol_name(prot), asp_name);
+			asp =
+			    osmo_ss7_asp_find_or_create(ss7, asp_name,
+							default_remote_port,
+							default_local_port,
+							prot);
+			talloc_free(asp_name);
+			if (!asp)
+				goto out_rt;
+			asp_created = true;
+			osmo_ss7_asp_peer_set_hosts(&asp->cfg.local, asp, &default_local_ip, 1);
+			osmo_ss7_asp_peer_set_hosts(&asp->cfg.remote, asp, &default_remote_ip, 1);
+			asp->simple_client_allocated = true;
+		}
 
 		osmo_ss7_as_add_asp(as, asp->cfg.name);
 	}
 
 	/* Ensure that the ASP we use is set to client mode. */
 	asp->cfg.is_server = false;
+	asp->cfg.role = OSMO_SS7_ASP_ROLE_ASP;
 
 	/* Restart ASP */
 	if (prot != OSMO_SS7_ASP_PROT_IPA)
@@ -442,9 +630,7 @@ osmo_sccp_simple_client_on_ss7_id(void *ctx, uint32_t ss7_id, const char *name,
 	LOGP(DLSCCP, LOGL_NOTICE, "%s: Using ASP instance %s\n", name,
 	     asp->cfg.name);
 
-	/* Allocate SCCP instance */
-	LOGP(DLSCCP, LOGL_NOTICE, "%s: Creating SCCP instance\n", name);
-	ss7->sccp = osmo_sccp_instance_create(ss7, NULL);
+	osmo_ss7_ensure_sccp(ss7);
 	if (!ss7->sccp)
 		goto out_asp;
 
@@ -484,9 +670,9 @@ osmo_sccp_simple_client(void *ctx, const char *name, uint32_t default_pc,
 {
 	/*! This is simplified version of osmo_sccp_simple_client_on_ss7_id().
 	 *  the only difference is that the ID of the CS7 instance will be
-	 *  set to 1 statically */
+	 *  set to 0 statically */
 
-	return osmo_sccp_simple_client_on_ss7_id(ctx, 1, name, default_pc, prot,
+	return osmo_sccp_simple_client_on_ss7_id(ctx, 0, name, default_pc, prot,
 						 default_local_port,
 						 default_local_ip,
 						 default_remote_port,
@@ -504,6 +690,7 @@ osmo_sccp_simple_server_on_ss7_id(void *ctx, uint32_t ss7_id, uint32_t pc,
 {
 	struct osmo_ss7_instance *ss7;
 	struct osmo_xua_server *xs;
+	int rc;
 
 	if (local_port < 0)
 		local_port = osmo_ss7_asp_protocol_port(prot);
@@ -518,8 +705,12 @@ osmo_sccp_simple_server_on_ss7_id(void *ctx, uint32_t ss7_id, uint32_t pc,
 	if (!xs)
 		goto out_ss7;
 
+	rc = osmo_ss7_xua_server_bind(xs);
+	if (rc < 0)
+		goto out_xs;
+
 	/* Allocate SCCP stack */
-	ss7->sccp = osmo_sccp_instance_create(ss7, NULL);
+	osmo_ss7_ensure_sccp(ss7);
 	if (!ss7->sccp)
 		goto out_xs;
 
@@ -538,7 +729,7 @@ osmo_sccp_simple_server(void *ctx, uint32_t pc,
 			enum osmo_ss7_asp_protocol prot, int local_port,
 			const char *local_ip)
 {
-	return osmo_sccp_simple_server_on_ss7_id(ctx, 1, pc, prot,
+	return osmo_sccp_simple_server_on_ss7_id(ctx, 0, pc, prot,
 					         local_port, local_ip);
 }
 
@@ -553,6 +744,7 @@ osmo_sccp_simple_server_add_clnt(struct osmo_sccp_instance *inst,
 	struct osmo_ss7_as *as;
 	struct osmo_ss7_route *rt;
 	struct osmo_ss7_asp *asp;
+	struct osmo_xua_server *oxs;
 	char *as_name, *asp_name;
 
 	if (local_port < 0)
@@ -568,7 +760,6 @@ osmo_sccp_simple_server_add_clnt(struct osmo_sccp_instance *inst,
 	as = osmo_ss7_as_find_or_create(ss7, as_name, prot);
 	if (!as)
 		goto out_strings;
-	talloc_free(as_name);
 
 	/* route only selected PC to the client */
 	rt = osmo_ss7_route_create(ss7->rtable_system, pc, 0xffff, as_name);
@@ -578,13 +769,26 @@ osmo_sccp_simple_server_add_clnt(struct osmo_sccp_instance *inst,
 	asp = osmo_ss7_asp_find_or_create(ss7, asp_name, remote_port, local_port, prot);
 	if (!asp)
 		goto out_rt;
+	oxs = osmo_ss7_xua_server_find(ss7, prot, local_port);
+	if (!oxs)
+		goto out_asp;
+	if (osmo_ss7_asp_peer_set_hosts(&asp->cfg.local, asp,
+					(const char* const*)oxs->cfg.local.host,
+					oxs->cfg.local.host_cnt) < 0)
+		goto out_asp;
+	if (osmo_ss7_asp_peer_add_host(&asp->cfg.remote, asp, remote_ip) < 0)
+		goto out_asp;
 	asp->cfg.is_server = true;
+	asp->cfg.role = OSMO_SS7_ASP_ROLE_SG;
 	osmo_ss7_as_add_asp(as, asp_name);
 	talloc_free(asp_name);
+	talloc_free(as_name);
 	osmo_ss7_asp_restart(asp);
 
 	return ss7->sccp;
 
+out_asp:
+	osmo_ss7_asp_destroy(asp);
 out_rt:
 	osmo_ss7_route_destroy(rt);
 out_as:
